/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.annotation.view.generic;

import static java.lang.Math.max;
import static org.fortiss.tooling.base.ui.annotation.editingsupport.EditingSupportFactory.createEditingSupport;
import static org.fortiss.tooling.base.ui.annotation.labelprovider.LabelProviderFactory.createLabelProvider;

import java.lang.reflect.Method;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.function.BiConsumer;
import java.util.stream.Stream;

import org.conqat.lib.commons.collections.Pair;
import org.eclipse.jface.viewers.ArrayContentProvider;
import org.eclipse.jface.viewers.CellLabelProvider;
import org.eclipse.jface.viewers.ColumnLabelProvider;
import org.eclipse.jface.viewers.ColumnViewerToolTipSupport;
import org.eclipse.jface.viewers.EditingSupport;
import org.eclipse.jface.viewers.IElementComparer;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.TableViewerColumn;
import org.eclipse.jface.viewers.ViewerCell;
import org.eclipse.jface.window.ToolTip;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.ControlEvent;
import org.eclipse.swt.events.ControlListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.MenuItem;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.TableItem;
import org.fortiss.tooling.base.annotation.AnnotationEntry;
import org.fortiss.tooling.base.annotation.IAnnotationValueService;
import org.fortiss.tooling.base.annotation.valueprovider.AnnotationInstSpec;
import org.fortiss.tooling.base.annotation.valueprovider.IAnnotationValueProvider;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.ui.annotation.editingsupport.EditingSupportFactory;
import org.fortiss.tooling.base.ui.annotation.editingsupport.ElementCommentEditingSupport;
import org.fortiss.tooling.base.ui.annotation.editingsupport.ElementNameEditingSupport;
import org.fortiss.tooling.base.ui.annotation.labelprovider.AnnotationLabelProvider;
import org.fortiss.tooling.base.ui.annotation.labelprovider.ElementCommentLabelProvider;
import org.fortiss.tooling.base.ui.annotation.labelprovider.ElementNameLabelProvider;
import org.fortiss.tooling.base.ui.annotation.view.AnnotationViewPartBase;
import org.fortiss.tooling.base.ui.annotation.view.generic.filter.AnnotationFilter;
import org.fortiss.tooling.base.ui.annotation.view.generic.filter.AnnotationFilterWidget;

/**
 * <p>
 * This class provides a generic table-based annotation view.
 * </p>
 * <p>
 * In order to support displaying and editing specific annotation types,
 * {@link EditingSupportFactory} must create an appropriate {@link EditingSupport} for the
 * corresponding {@link IAnnotationValueProvider} base classes.
 * </p>
 * 
 * @author eder, diewald, barner
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: E73C8AEA023803F8E01B7C576DDAC5A4
 */
public class GenericAnnotationView extends AnnotationViewPartBase {
	/** Root composite of {@link GenericAnnotationView}. */
	private Composite rootComposite;

	/** The table viewer used to implement the view. */
	private TableViewer tableViewer;

	/** Caches the currently selected item in the table to make it available to listeners. */
	IStructuredSelection selectedTableItem;

	/** Associates the columns of the annotation view with the providers responsible for them. */
	Map<AnnotationInstSpec<? extends IAnnotatedSpecification>, IAnnotationValueProvider<IAnnotatedSpecification>> annotationCtxEntries =
			new HashMap<>();

	/** Row and column filter for annotations. */
	private AnnotationFilter annotationFilter;

	/** Widget for filter options. */
	private AnnotationFilterWidget annotationFilterWidget;

	/**
	 * Last set of {@link AnnotationEntry}s used to {@link #update(Collection)} the
	 * {@link GenericAnnotationView}, i.e., during the construction of the, or because of a model
	 * change. The reference is required to update view after a change of the column (annotation)
	 * filter.
	 */
	private Collection<AnnotationEntry> lastAnnotationEntries;

	/** Number of fixed columns shown for every model element (e.g., name, comment). */
	private int fixedColumnCount;

	/**
	 * Map used to preserve column width of annotation columns during update of the view (fixed
	 * columns are not deleted hence caching the width is not necessary).
	 */
	private Map<String, Integer> columnWidthCache;

	/** Default width of columns. */
	private static final int COLUMN_DEFAULT_WIDTH = 125;

	/** Default height of rows. */
	private int defaultRowHeight;

	/**
	 * Sets a fixed {@link AnnotationFilter}.
	 * <ul>
	 * <li>If the {@link AnnotationFilter} is non-{@code null}, it sets the filter and disables the
	 * {@link AnnotationFilterWidget}. For instance, this can be used to embed / extend the
	 * {@link GenericAnnotationView} into a specialized view dedicated to a particular model type.</li>
	 * <li>Otherwise, this method reinitializes the {@link AnnotationFilterWidget}.</li>
	 * </ul>
	 */
	public void setFixedAnnotationFilter(AnnotationFilter annotationFilter) {

		// Remove old table row filter
		tableViewer.removeFilter(this.annotationFilter);

		if(annotationFilter != null) {
			// Externally created filter -> remove filter option GUI
			this.annotationFilter = annotationFilter;
			if(annotationFilterWidget != null) {
				annotationFilterWidget.dispose();
				annotationFilterWidget = null;
			}
		} else {
			// No externally created filter -> setup filter option GUI
			this.annotationFilter = new AnnotationFilter();
			setupAnnotationFilterWidget();
		}

		// Add new table row filter
		tableViewer.addFilter(this.annotationFilter);

		// Update layout
		rootComposite.layout(true, true);
	}

	/** Returns the view's {@link AnnotationFilter}. */
	public AnnotationFilter getAnnotationFilter() {
		return annotationFilter;
	}

	/** {@inheritDoc} */
	@Override
	public void update() {
		// Full update, i.e. including reconstructions of columns is always required due to the
		// following scenario:
		// - Filter is set for annotation names.
		// - Filter scope is changed to component names -> annotation filter needs to be undone
		// - Filter is based on annotation entry type
		// - Filter is set to current level / current sub model
		// - Update of view as been disabled using setUpdateEnable(boolean) is re-enabled now
		if(lastAnnotationEntries != null) {
			update(lastAnnotationEntries);
		}
	}

	/** {@inheritDoc} */
	@Override
	protected void update(Collection<AnnotationEntry> annotationEntries) {
		if(!isUpdateEnabled()) {
			return;
		}

		// Update row filter with currently selected element
		annotationFilter.setCurrentySelectedModelElement(getCurrentlySelectedObject());
		boolean contentRequiresUpdate =
				lastAnnotationEntries == null || !annotationEntries.equals(lastAnnotationEntries);

		if(annotationFilter.annotationViewQueryUpdateRequired() || contentRequiresUpdate) {
			// Update filter widget
			if(annotationFilterWidget != null) {
				annotationFilterWidget.updateAnnotationTypeFilterOptionsComboBox(annotationEntries,
						this);
			}

			// Remember last annotation entries (required to re-draw table after changing the
			// column/annotation filter
			lastAnnotationEntries = annotationEntries;

			// Minimize flickering while updating the table
			tableViewer.getTable().setRedraw(false);

			// Input must be set before the construction of the columns, since the editing support
			// for a given cell might need to query the contents of other model elements (e.g., in
			// order to aggregate a set of admissible choices for its value).
			tableViewer.setInput(annotationEntries);

			// Sorted set of {@link ColumnHandle} used to instantiate this {@link
			// GenericAnnotationView}'s columns in the right order.
			Set<ColumnHandle<?>> sortedColumnHandles = new TreeSet<ColumnHandle<?>>();

			// Dispose all columns of the table viewer, except the leading fixed columns, which e.g.
			// hold the model elements' names and comments.
			while(tableViewer.getTable().getColumnCount() > fixedColumnCount) {
				TableColumn column = tableViewer.getTable().getColumn(fixedColumnCount);
				column.dispose();
			}

			// Aggregate required columns. Column order is defined by ColumnHandle.compareTo().
			for(AnnotationEntry entry : annotationEntries) {
				for(IAnnotatedSpecification spec : entry.getSpecificationsList()) {
					// Create new column for each instance of the current multi-instance annotation
					Class<? extends IAnnotatedSpecification> specClass = spec.getClass();
					boolean isSingleInstanceAnnotation = entry.getInstanceKeys(specClass).isEmpty();
					boolean allowsDynamicInstances =
							entry.allowsDynamicAnnotationInstances(specClass);
					if(allowsDynamicInstances || !isSingleInstanceAnnotation) {

						for(String instanceKey : entry.getInstanceKeys(specClass)) {
							ColumnHandle<IAnnotatedSpecification> columnHandle =
									new ColumnHandle<IAnnotatedSpecification>(entry, spec,
											instanceKey, false);
							if(annotationFilter.passesColumnFilter(columnHandle)) {
								sortedColumnHandles.add(columnHandle);
							}
						}
					}

					// Create a new column for
					// - ordinary (single-instance) annotations
					// - "create" column for multi-instance annotations for which instances can be
					// created dynamically
					if(allowsDynamicInstances || isSingleInstanceAnnotation) {
						ColumnHandle<IAnnotatedSpecification> columnHandle =
								new ColumnHandle<IAnnotatedSpecification>(entry, spec, null,
										allowsDynamicInstances);

						if(annotationFilter.passesColumnFilter(columnHandle)) {
							sortedColumnHandles.add(columnHandle);
						}
					}
				}
			}

			// Instantiate columns
			for(ColumnHandle<?> columnHandle : sortedColumnHandles) {
				createAnnotationColumn(columnHandle);
			}

			tableViewer.getTable().setRedraw(true);

			// After contents change, restore row height. It will be adjusted to the new content
			// using the TableViewer's SWT.{MeasureItem, PaintItem, EraseItem} listeners installed
			// in addRowHeightListener(). This needs to be done here since in the above listeners,
			// the table row height can only grow and may not be decreased.
			// See <https://bugs.eclipse.org/bugs/show_bug.cgi?id=154341>.
			setTableItemHeight(tableViewer.getTable(), defaultRowHeight);
		}

		// Update the view
		tableViewer.refresh();
	}

	/** {@inheritDoc} */
	@Override
	protected void update(IModelElement modelElement) {
		// Updates the contents of the table
		super.update(modelElement);

		// ensure that selected model element is visible in annotation view
		AnnotationEntry entry =
				IAnnotationValueService.getInstance().getAnnotationEntry(modelElement);
		assert (entry.getModelElement() == modelElement);

		tableViewer.reveal(entry);
	}

	/** Initializes the {@link AnnotationFilterWidget}. */
	private void setupAnnotationFilterWidget() {
		annotationFilterWidget = new AnnotationFilterWidget(rootComposite, SWT.NULL, this);
		annotationFilterWidget.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 1, 1));
	}

	/**
	 * Sets the item height of the given {@link Table} to the specified value.
	 *
	 * <b>The underlying method {@code Table#setItemHeight} is not exposed. This wrapper method
	 * should only called during a full {@link #update(Collection)} of the table's contents.</b>>
	 */
	private void setTableItemHeight(Table table, int height) {
		try {
			Method method = table.getClass().getDeclaredMethod("setItemHeight", int.class);
			if(method != null) {
				boolean accessible = method.isAccessible();
				method.setAccessible(true);
				method.invoke(table, height);
				method.setAccessible(accessible);
			}
		} catch(Exception e) {
			return;
		}
	}

	/**
	 * Installs a listener to the given {@link Table} that adjust the row height to to the current
	 * contents.
	 */
	private void addRowHeightListener(Table table) {

		/*
		 * NB: MeasureItem, PaintItem and EraseItem are called repeatedly and are hence
		 * performance critical.
		 */
		Listener rowHeightListener = new Listener() {
			@Override
			public void handleEvent(Event event) {
				switch(event.type) {
					case SWT.MeasureItem: {
						TableItem item = (TableItem)event.item;
						String text = item.getText(event.index);
						Point textSize = event.gc.textExtent(text);
						event.width = textSize.x;
						event.height = Math.max(event.height, textSize.y);

						Image image = item.getImage(event.index);
						if(image != null) {
							Rectangle imageSize = image.getBounds();
							event.width = max(event.width, imageSize.x);
							event.height = max(event.height, imageSize.y);
						}
						break;
					}
					case SWT.PaintItem: {
						TableItem item = (TableItem)event.item;

						// Determine maximum content height (considering text and image contents),
						// and derive resulting vertical offset
						String text = item.getText(event.index);
						int contentHeight = event.gc.textExtent(text).y;

						Image image = item.getImage(event.index);
						if(image != null) {
							contentHeight = max(contentHeight, image.getBounds().y);
						}
						int offset = max(0, (event.height - contentHeight) / 2);

						// Render content
						event.gc.drawText(text, event.x, event.y + offset, true);
						if(image != null) {
							event.gc.drawImage(image, event.x, event.y + offset);
						}

						break;
					}
					case SWT.EraseItem: {
						event.detail &= ~SWT.FOREGROUND;
						break;
					}
				}
			}
		};
		table.addListener(SWT.MeasureItem, rowHeightListener);
		table.addListener(SWT.PaintItem, rowHeightListener);
		table.addListener(SWT.EraseItem, rowHeightListener);
	}

	/**
	 * Constructs a {@link Listener} such that a context menu is created upon a right-click on
	 * entries in the Annotation View.
	 */
	private void addContextMenuListener(Table table) {
		table.addListener(SWT.MouseDown, new Listener() {
			@Override
			public void handleEvent(Event event) {
				TableItem[] selection = table.getSelection();
				if(selection.length != 0 && (event.button == 3)) {

					Menu contextMenu = new Menu(table);
					table.setMenu(contextMenu);

					ViewerCell cell = tableViewer.getCell(new Point(event.x, event.y));
					if(cell == null) {
						return;
					}
					CellLabelProvider labelProv =
							tableViewer.getLabelProvider(cell.getColumnIndex());
					if(labelProv instanceof AnnotationLabelProvider) {
						@SuppressWarnings("unchecked") Class<IAnnotatedSpecification> annotationType =
								(Class<IAnnotatedSpecification>)((AnnotationLabelProvider)labelProv)
										.getAnnotationType();
						String instanceKey = ((AnnotationLabelProvider)labelProv).getInstanceKey();
						AnnotationInstSpec<IAnnotatedSpecification> spec =
								new AnnotationInstSpec<IAnnotatedSpecification>(annotationType,
										instanceKey);
						addContextMenuEntries(contextMenu, annotationCtxEntries.get(spec), spec);
					}
					contextMenu.setVisible(true);
				}
			}
		});
	}

	/**
	 * Adds a {@link Listener} to the given {@link Menu} entry such that the given
	 * {@link BiConsumer} is executed upon selection.
	 */
	private
			<T extends IAnnotatedSpecification, S extends IModelElement>
			void
			addContextMenuEntryListener(
					MenuItem entry,
					BiConsumer<AnnotationInstSpec<IAnnotatedSpecification>, AnnotationEntry> action,
					AnnotationInstSpec<IAnnotatedSpecification> specificationType) {
		entry.addSelectionListener(new SelectionAdapter() {
			AnnotationInstSpec<IAnnotatedSpecification> spec = specificationType;

			@Override
			public void widgetSelected(SelectionEvent e) {
				if(selectedTableItem.getFirstElement() instanceof AnnotationEntry) {
					AnnotationEntry entry = (AnnotationEntry)selectedTableItem.getFirstElement();
					boolean updateEnabled = isUpdateEnabled();
					setUpdateEnabled(false);
					action.accept(spec, entry);
					setUpdateEnabled(updateEnabled);
				}
			}
		});
	}

	/**
	 * Adds context menu entries to the given {@link Menu} whose name and associated action is
	 * determined by the {@link IAnnotationValueProvider}.
	 */
	private void addContextMenuEntries(Menu contextMenu,
			IAnnotationValueProvider<IAnnotatedSpecification> valueProvider,
			AnnotationInstSpec<IAnnotatedSpecification> annotationClass) {
		final Stream<AnnotationEntry> annotationEntriesMatchingFilter =
				getLastAnnotationEntries().stream()
						.filter(e -> annotationFilter.passesRowFilter(e));
		for(Pair<String, BiConsumer<AnnotationInstSpec<IAnnotatedSpecification>, AnnotationEntry>> ctxMenuEntry : valueProvider
				.getContextMenuEntries(annotationEntriesMatchingFilter)) {
			MenuItem mItem = new MenuItem(contextMenu, SWT.None);
			mItem.setText(ctxMenuEntry.getFirst());
			addContextMenuEntryListener(mItem, ctxMenuEntry.getSecond(), annotationClass);
		}
	}

	/** {@inheritDoc} */
	@Override
	public void createPartControl(Composite parent) {
		super.createPartControl(parent);
		rootComposite = new Composite(parent, SWT.NULL);
		rootComposite.setLayout(new GridLayout(1, false));

		tableViewer = new TableViewer(rootComposite, SWT.BORDER | SWT.FULL_SELECTION);
		Table table = tableViewer.getTable();
		table.setHeaderVisible(true);
		table.setLinesVisible(true);
		table.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true, 1, 1));

		// Listener to adjust row height to contents
		addRowHeightListener(table);

		// Preserve default row-height (used to restore it during contents updates in
		// update(Collection<AnnotationEntry>)).
		defaultRowHeight = table.getItemHeight();

		ColumnViewerToolTipSupport.enableFor(tableViewer, ToolTip.NO_RECREATE);

		// Install layout for Table in order to ensure that it claims all available vertical space
		// in case the
		// filter option GUI is removed (see setFixedAnnotationFilter()).
		table.setLayout(new FillLayout(SWT.VERTICAL));

		fixedColumnCount = createFixedModelElementColumns();
		columnWidthCache = new HashMap<String, Integer>();

		tableViewer.setComparator(new HierarchicalNameComparator());

		tableViewer.setComparer(new IElementComparer() {

			@Override
			public int hashCode(Object element) {
				if(element instanceof AnnotationEntry) {
					return ((AnnotationEntry)element).getModelElement().hashCode();
				}
				return element.hashCode();
			}

			@Override
			public boolean equals(Object a, Object b) {
				if(a instanceof AnnotationEntry && b instanceof AnnotationEntry) {
					return ((AnnotationEntry)a).getModelElement() == ((AnnotationEntry)b)
							.getModelElement();
				}
				return a.equals(b);
			}
		});

		tableViewer.addSelectionChangedListener(new ISelectionChangedListener() {
			@Override
			public void selectionChanged(final SelectionChangedEvent event) {
				selectedTableItem = (IStructuredSelection)event.getSelection();
			}
		});

		addContextMenuListener(table);

		// Rows (= components) are filtered using a specialized ViewerFilter.
		// Since the data model is an array of AnnotationEntries, columns (= annotations)
		// cannot be filtered using such a filter, and require the reconstruction of the
		// table, with an implementation of the filter in update(Set<AnnotationEntry>).
		annotationFilter = new AnnotationFilter();
		tableViewer.addFilter(annotationFilter);

		// Add filter widget
		setupAnnotationFilterWidget();
	}

	/** Creates a fixed (leading) column to be shown for every model element. */
	private void createFixedColumn(String label, ColumnLabelProvider labelProvider,
			EditingSupport editingSupport) {
		TableViewerColumn column = new TableViewerColumn(tableViewer, SWT.NONE);
		column.getColumn().setText(label);
		column.getColumn().setWidth(COLUMN_DEFAULT_WIDTH);

		tableViewer.setContentProvider(new ArrayContentProvider());
		column.setLabelProvider(labelProvider);
		column.setEditingSupport(editingSupport);
	}

	/**
	 * Creates the (leading) columns which displays the model elements' names and comments.
	 * 
	 * @returns Number of leading columns.
	 * */
	protected int createFixedModelElementColumns() {
		createFixedColumn("Model Element", new ElementNameLabelProvider(this),
				new ElementNameEditingSupport(tableViewer));

		createFixedColumn("Comment", new ElementCommentLabelProvider(this),
				new ElementCommentEditingSupport(tableViewer));

		return 2;
	}

	/**
	 * Creates a column in the table viewer displaying the annotation identified by the given
	 * {@link ColumnHandle}.
	 */
	protected void createAnnotationColumn(
			ColumnHandle<? extends IAnnotatedSpecification> columnHandle) {

		TableColumn tableColumn = null;
		if(columnHandle.isCreateColumn()) {
			// Create column that contains a button for adding new instances of the
			// current annotation type.

			CreateAnnotationInstanceColumn column =
					new CreateAnnotationInstanceColumn(tableViewer, SWT.NONE,
							columnHandle.getAnnotatedSpecification(), columnHandle.getEntry());
			tableColumn = column.getColumn();
		} else {
			@SuppressWarnings("unchecked") Class<IAnnotatedSpecification> annotationClass =
					(Class<IAnnotatedSpecification>)columnHandle.getAnnotatedSpecification()
							.getClass();

			// Add new column
			try {
				@SuppressWarnings("unchecked") TableViewerColumn column =
						createAnnotationTableViewerColumn(
								(ColumnHandle<IAnnotatedSpecification>)columnHandle,
								annotationClass);

				tableColumn = column.getColumn();
				String specName = columnHandle.getColumnName();
				tableColumn.setText(specName);
			} catch(Exception e) {
				System.out.println("Failed to create table viewer column for annotation class " +
						annotationClass.getCanonicalName() +
						((e.getMessage() != null) ? ": " + e.getMessage() : "."));
			}
		}

		if(tableColumn != null) {
			tableColumn.addControlListener(new ControlListener() {
				@Override
				public void controlResized(ControlEvent e) {
					TableColumn tc = (TableColumn)e.getSource();
					columnWidthCache.put(tc.getText(), tc.getWidth());
				}

				@Override
				public void controlMoved(ControlEvent e) {
					// Nothing to do
				}
			});
			Integer width = columnWidthCache.get(tableColumn.getText());
			if(width != null) {
				// If possible, use cached with (possibly adjusted by the user)
				tableColumn.setWidth(width);
			} else {
				// Otherwise, calculate width for new columns
				tableColumn.pack();
				width = tableColumn.getWidth();
				if(width < COLUMN_DEFAULT_WIDTH) {
					tableColumn.setWidth(COLUMN_DEFAULT_WIDTH);
				} else if(width > 1.2f * COLUMN_DEFAULT_WIDTH) {
					tableColumn.setWidth(Math.round(1.2f * COLUMN_DEFAULT_WIDTH));
				}
			}
		}
	}

	/**
	 * Creates a {@link TableViewerColumn}, i.e., instantiate matching {@link EditingSupport} and
	 * {@link LabelProvider} for the given annotation.
	 */
	private
			<T extends IAnnotatedSpecification>
			TableViewerColumn
			createAnnotationTableViewerColumn(ColumnHandle<T> columnHandle, Class<T> annotationClass)
					throws Exception {
		TableViewerColumn column = new TableViewerColumn(tableViewer, SWT.NONE);

		// Have the matching EditingSupport created for the current annotated specification.
		AnnotationEntry entry = columnHandle.getEntry();
		IAnnotationValueProvider<IAnnotatedSpecification> valueProvider =
				entry.getAnnotationValueProvider(annotationClass);
		T specification = entry.getSpecification(annotationClass);
		String instanceKey = columnHandle.getInstanceKey();

		EditingSupport editingSupport =
				createEditingSupport(valueProvider, tableViewer, annotationClass, specification,
						instanceKey);

		ColumnLabelProvider columnLabelProvider =
				createLabelProvider(valueProvider, this, tableViewer, editingSupport,
						annotationClass, specification, instanceKey);

		column.setEditingSupport(editingSupport);
		column.setLabelProvider(columnLabelProvider);

		annotationCtxEntries.put(columnHandle, valueProvider);
		// addContextMenuEntries(valueProvider, annotationClass);

		return column;
	}

	/** {@inheritDoc} */
	@Override
	public void setFocus() {
		// nothing to do
	}

	/** {@inheritDoc} */
	@Override
	public synchronized void dispose() {
		tableViewer.getTable().dispose();
		super.dispose();
	}

	/** Returns {@link #lastAnnotationEntries}. */
	public Collection<AnnotationEntry> getLastAnnotationEntries() {
		return lastAnnotationEntries;
	}
}
