/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.annotation.valueprovider;

import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.function.BiConsumer;
import java.util.stream.Stream;

import org.conqat.lib.commons.collections.Pair;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.fortiss.tooling.base.annotation.AnnotationEntry;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.kernel.service.base.IEObjectAware;

/**
 * <p>
 * Value provider interface for {@link IAnnotatedSpecification}s that establishes the link between
 * the model (data storage) and the UI (data exchange and conversion). Concrete annotations need to
 * provide specific implementations of this interface to perform the data exchange and conversion,
 * and to provide the required factory methods.
 * </p>
 * Instance keys:
 * <ul>
 * <li>Since multiple instances can exist for a single {@link IAnnotatedSpecification} (dynamic
 * instantiation by the user (see {@link DynamicInstanceAnnotationValueProviderBase}), <b>instance
 * keys</b> are used to differentiate the different instances.</li>
 * <li>Additionally, keys are used to differentiate the different annotations that can be hosted by
 * a single {@link IAnnotatedSpecification}. A typical use case is a single
 * {@link EStructuralFeatureValueProviderBase} that supports multiple (or all)
 * {@link EStructuralFeature} provided by single annotation class.</li>
 * <li>In case the annotation only provides a single value, the {@link #DEFAULT_KEY} is used. For
 * convenience, the method interface provides wrapper methods that omits the {@code instanceKey}
 * parameter and wrap the corresponding generic method using the {@link #DEFAULT_KEY}).</li>
 * </ul>
 * </p>
 * 
 * <p>
 * The interface / some of its methods take two template parameters:
 * 
 * <ul>
 * <li><tt>T</tt>: Meta-model type of the specification, i.e., the particular subclass of
 * {@link IAnnotatedSpecification} supported by the respective implementation of this
 * {@link IAnnotationValueProvider} interface.</li>
 * <li><tt>U</tt>: Data type used in the UI to support the respective annotation. For <tt>U</tt>, a
 * reasonable {@code toString()} implementation must be provided that is used by the framework to
 * render the value of the annotation in text-based editing support UI controls.</li>
 * </ul>
 * </p>
 * 
 * @author eder, diewald, barner
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 35E646C624C6EEEF82A16AC2EC255A4D
 */
public interface IAnnotationValueProvider<T extends IAnnotatedSpecification> extends
		IEObjectAware<IModelElement> {

	/**
	 * Default key used in case this {@link IAnnotationValueProvider} is used to manage a single
	 * instance key only.
	 */
	public static String DEFAULT_KEY = null;

	/** Returns the name of the annotation (e.g., used as labels in annotation views). */
	public String getAnnotationName(T specification);

	/**
	 * Sets a new name for the annotation (e.g., used as labels in annotation views). The return
	 * value indicates if the rename operation was successful.
	 */
	public boolean setAnnotationName(String name, T specification);

	/** Predicate if the the given annotation should be hidden from any view. */
	public boolean isHidden(T specification);

	/** Predicate if the given annotation can be edited. */
	public boolean canEdit(T specification);

	/** Predicate if the given annotation can be edited. */
	public boolean canEdit(T specification, String instanceKey);

	/** Returns the value of the annotation. */
	public <U> U getAnnotationValue(T specification) throws Exception;

	/** Returns the value of a multi-instance annotation's instance denoted by {@code instanceKey}. */
	public <U> U getAnnotationValue(T specification, String instanceKey) throws Exception;

	/**
	 * Sets a new value for the annotation.
	 * 
	 * @throws Exception
	 *             is thrown if the conversion from the UI internal value representation ({@code U})
	 *             to the respective specialization of {@link IAnnotatedSpecification} has failed,
	 *             or if this operation is not available.
	 */
	public <U> void setAnnotationValue(U value, T specification) throws Exception;

	/**
	 * Sets a new value in the annotation from a {@link String} initializer.
	 * 
	 * @throws Exception
	 *             is thrown if the conversion from {@link String} to the respective specialization
	 *             of {@link IAnnotatedSpecification} has failed, or if this operation is not
	 *             available.
	 */
	public void setAnnotationValue(String value, T specification) throws Exception;

	/**
	 * Sets a new value for a multi-instance annotation's instance denoted by {@code instanceKey}.
	 * 
	 * @throws Exception
	 *             is thrown if the conversion from the UI internal value representation ({@code U})
	 *             to the respective specialization of {@link IAnnotatedSpecification} has failed,
	 *             or if this operation is not available.
	 */
	public <U> void setAnnotationValue(U value, T specification, String instanceKey)
			throws Exception;

	/**
	 * Sets a new value for a multi-instance annotation's instance denoted by {@code instanceKey}
	 * from a {@link String} initializer.
	 * 
	 * @throws Exception
	 *             is thrown if the conversion from {@link String} to the respective specialization
	 *             of {@link IAnnotatedSpecification} has failed, or if this operation is not
	 *             available.
	 */
	public void setAnnotationValue(String value, T specification, String instanceKey)
			throws Exception;

	/**
	 * Factory method that creates a new {@link IAnnotatedSpecification} specialization of type T to
	 * the given model {@code element} and returns a reference to it.
	 * 
	 * In case the annotation already exists in the model, this method does not create a new object,
	 * but returns the existing one.
	 * 
	 * @throws Exception
	 *             is thrown if no existing element could be retrieved, and the creation of a new
	 *             element has failed.
	 */
	public T getAnnotatedSpecificationForModelElement(IModelElement element) throws Exception;

	/**
	 * Returns the concrete specialization {@code T} of {@link IAnnotatedSpecification} supported by
	 * this specialization of {@link IAnnotationValueProvider}.
	 */
	public Class<T> getAnnotationClazz();

	/**
	 * Returns a {@link List} of model element types which will not have the
	 * {@link IAnnotatedSpecification} managed by this {@link IAnnotationValueProvider}, refining
	 * the registration of the respective {@link IAnnotatedSpecification} in the
	 * {@code org.fortiss.tooling.base.ui.annotation} extension point.
	 */
	public List<Class<? extends EObject>> excludeModelElementTypeFromAnnotatedSpecification();

	/**
	 * Returns the instance keys supported by this {@link IAnnotationValueProvider}.
	 * {@link IAnnotationValueProvider}s can either define a static {@link List}, or enable the
	 * dynamic creation of instance keys using {@link #allowsDynamicAnnotationInstances()}. In the
	 * latter case, this method returns current list of dynamic annotation instances.
	 */
	public Collection<String> getInstanceKeys(T specification);

	/**
	 * <p>
	 * Predicate whether the {@link IAnnotatedSpecification} managed by this
	 * {@link IAnnotationValueProvider} contains a value type allowing the storage of dynamically
	 * created instances of the annotation (like a list of strings).
	 * </p>
	 * <p>
	 * Static instances can be implemented by overriding
	 * {@link #getInstanceKeys(IAnnotatedSpecification)} and returning the {@link List} of
	 * admissible instance keys.
	 * </p>
	 */
	public boolean allowsDynamicAnnotationInstances();

	/**
	 * <p>
	 * Returns a list of String <--> {@link BiConsumer} {@link Pair}s that define the context menu
	 * entries that are available for specific entries of an {@link IAnnotatedSpecification}.
	 * </p>
	 * <p>
	 * This list is used in the view part of the annotations to provide actions for concrete
	 * {@link AnnotationEntry}s that can be triggered via the context menu in the annotation view.
	 * The action is applied to the subset of {@link AnnotationEntry}s provided by the underlying
	 * annotation view (typically, the set of visible annotations that match the current filters).
	 * </p>
	 * 
	 * @param annotationEntries
	 *            {@link Stream} of {@link AnnotationEntry}s for which the actions should be
	 *            offered.
	 * 
	 * @return List of String <--> {@link BiConsumer} {@link Pair}s that define the context menu
	 *         entries that are available for specific entries of an {@link IAnnotatedSpecification}
	 *         .
	 */
	default List<Pair<String, BiConsumer<AnnotationInstSpec<T>, AnnotationEntry>>>
			getContextMenuEntries(Stream<AnnotationEntry> annotationEntries) {
		return Collections.emptyList();
	}
}
