/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.utils;

import java.util.NoSuchElementException;
import java.util.function.Consumer;

import org.eclipse.emf.common.util.EList;
import org.fortiss.tooling.kernel.extension.IConstraint;
import org.fortiss.tooling.kernel.model.constraints.ConstrainedWithChecksum;
import org.fortiss.tooling.kernel.model.constraints.ConstraintInstance;
import org.fortiss.tooling.kernel.model.constraints.ConstraintsFactory;
import org.fortiss.tooling.kernel.model.constraints.ErrorConstraintInstanceStatus;
import org.fortiss.tooling.kernel.model.constraints.FailedConstraintInstanceStatus;
import org.fortiss.tooling.kernel.model.constraints.IConstrained;
import org.fortiss.tooling.kernel.model.constraints.IConstraintInstanceContainer;
import org.fortiss.tooling.kernel.model.constraints.IConstraintInstanceStatus;
import org.fortiss.tooling.kernel.model.constraints.OutdatedConstraintInstanceStatus;
import org.fortiss.tooling.kernel.model.constraints.SuccessConstraintInstanceStatus;
import org.fortiss.tooling.kernel.service.IConstraintService;
import org.fortiss.tooling.kernel.service.IConstraintService.IFix;

/**
 * Utility methods for dealing with constraints.
 * 
 * @author aravantinos
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 115825B5DDF0458D48AC3BEA8D4CBF97
 */
public class ConstraintsUtils {

	/**
	 * Returns the container of <code>ci</code>. Assumes that it is of type
	 * {@link IConstraintInstanceContainer}!
	 */
	public static IConstraintInstanceContainer
			getConstraintInstanceContainer(ConstraintInstance ci) {
		return (IConstraintInstanceContainer)ci.eContainer();
	}

	/**
	 * Defines <code>constrained</code> as the <code>i</code>-th constrained model element of
	 * <code>ci</code>. Leaves the other elements as they are if they exist, fill them with
	 * <code>null</code> till <code>i</code> otherwise.
	 */
	public static void
			setConstrainedElement(ConstraintInstance ci, IConstrained constrained, int i) {
		EList<ConstrainedWithChecksum> cwcs = ci.getConstrainedsWithChecksum();
		if(cwcs.size() <= i) {
			for(int j = cwcs.size(); j <= i; j++) {
				cwcs.add(j, ConstraintsFactory.eINSTANCE.createConstrainedWithChecksum());
			}
		}
		if(!constrained.getConstraintInstances().contains(ci)) {
			constrained.getConstraintInstances().add(ci);
		}
		ConstrainedWithChecksum cwc = cwcs.get(i);
		cwc.setConstrained(constrained);
		cwc.setChecksum(null);
	}

	/** @return the first constraint instance of type <code>cstr</code> in <code>c</code>. */
	public static ConstraintInstance getConstraintInstanceOfType(IConstrained c, IConstraint cstr) {
		return getConstraintInstanceOfType(c, cstr.getClass());
	}

	/** @return the first constraint instance of type <code>cstr</code> in <code>c</code>. */
	public static ConstraintInstance getConstraintInstanceOfType(IConstrained c,
			Class<? extends IConstraint> cstrClass) {
		try {
			String name = cstrClass.getName();
			return c.getConstraintInstances().stream()
					.filter(x -> name.equals(x.getConstraintName())).findFirst().get();
		} catch(NoSuchElementException e) {
			return null;
		}
	}

	/** Exception which embeds a constraint error status. */
	public static class ErrorEmbeddingException extends RuntimeException {

		/** The embedded error. */
		public ErrorConstraintInstanceStatus error;

		/** Constructor. */
		public ErrorEmbeddingException(ErrorConstraintInstanceStatus error) {
			super();
			this.error = error;
		}
	}

	/** @return A "success" status. */
	public static SuccessConstraintInstanceStatus createSuccessStatus() {
		return ConstraintsFactory.eINSTANCE.createSuccessConstraintInstanceStatus();
	}

	/** @return A "outdated" status for <code>ci</code>. */
	public static OutdatedConstraintInstanceStatus createOutdatedStatus(ConstraintInstance ci) {
		OutdatedConstraintInstanceStatus status;
		if(ci.getStatus() instanceof OutdatedConstraintInstanceStatus) {
			status = (OutdatedConstraintInstanceStatus)ci.getStatus();
		} else {
			status = ConstraintsFactory.eINSTANCE.createOutdatedConstraintInstanceStatus();
			ci.setStatus(status);
		}
		return status;
	}

	/** @return A "outdated" status. */
	public static OutdatedConstraintInstanceStatus createOutdatedStatus() {
		return ConstraintsFactory.eINSTANCE.createOutdatedConstraintInstanceStatus();
	}

	/** @return A "fail" status. */
	public static FailedConstraintInstanceStatus createFailStatus() {
		return ConstraintsFactory.eINSTANCE.createFailedConstraintInstanceStatus();
	}

	/** @return An "error" status. */
	public static ErrorConstraintInstanceStatus createErrorStatus() {
		return ConstraintsFactory.eINSTANCE.createErrorConstraintInstanceStatus();
	}

	/** @return the first element constrained by <code>ci</code> */
	public static IConstrained getFirstConstrained(ConstraintInstance ci) {
		try {
			// get(0) because the constraint instance shall contain at least one constrained element
			// (responsibility of the caller!) and because we want the first element.
			return ci.getConstrainedsWithChecksum().get(0).getConstrained();
		} catch(IndexOutOfBoundsException e) {
			return null;
		}
	}

	/** Sets the first constrained element of <code>ci</code> to <code>constrained</code>. */
	public static void setFirstConstrained(ConstraintInstance ci, IConstrained constrained) {
		setConstrainedElement(ci, constrained, 0);
	}

	/** @return the second element constrained by <code>ci</code> */
	public static IConstrained getSecondConstrained(ConstraintInstance ci) {
		try {
			// get(1) because the constraint instance shall contain at least one constrained element
			// (responsibility of the caller!) and because we want the first element.
			return ci.getConstrainedsWithChecksum().get(1).getConstrained();
		} catch(IndexOutOfBoundsException e) {
			return null;
		}
	}

	/** Sets the second constrained element of <code>ci</code> to <code>constrained</code>. */
	public static void setSecondConstrained(ConstraintInstance ci, IConstrained constrained) {
		setConstrainedElement(ci, constrained, 1);
	}

	/**
	 * @return the element constrained by <code>ci</code>.
	 *         Use this function if your constraint constrains only one element!
	 */
	public static IConstrained getConstrained(ConstraintInstance ci) {
		return getFirstConstrained(ci);
	}

	/**
	 * Sets the constrained element of <code>ci</code> to <code>constrained</code>.
	 * Use this function if your constraint constrains only one element!
	 */
	public static void setConstrained(ConstraintInstance ci, IConstrained constrained) {
		setConstrainedElement(ci, constrained, 0);
	}

	/**
	 * @return a constraint instance of type <code>cstrClass</code> constraining
	 *         <code>constrained</code>
	 */
	public static ConstraintInstance createConstraintInstance(
			Class<? extends IConstraint> cstrClass, IConstrained constrained) {
		ConstraintInstance ci = ConstraintsFactory.eINSTANCE.createConstraintInstance();
		ci.setConstraintName(cstrClass.getName());
		setConstrained(ci, constrained);
		createOutdatedStatus(ci);
		return ci;
	}

	/**
	 * @return a constraint instance of type <code>cstrClass</code> constraining
	 *         <code>constrained1</code> and <code>constrained2</code>.
	 */
	public static ConstraintInstance createConstraintInstance(
			Class<? extends IConstraint> cstrClass, IConstrained constrained1,
			IConstrained constrained2) {
		ConstraintInstance c = ConstraintsFactory.eINSTANCE.createConstraintInstance();
		c.setConstraintName(IConstraintService.getInstance().getName(cstrClass));
		ConstraintsUtils.setFirstConstrained(c, constrained1);
		ConstraintsUtils.setSecondConstrained(c, constrained2);
		createOutdatedStatus(c);
		return c;
	}

	/** @return a successful status if <code>b</code> is true, a fail status otherwise */
	static public IConstraintInstanceStatus successIfTrue(boolean b) {
		return b ? createSuccessStatus() : createFailStatus();
	}

	/** Base for fixes. */
	public static class FixBase implements IFix {
		/** The description. */
		private String description;

		/** The fix itself. */
		private Consumer<IConstraintInstanceStatus> fix;

		/** Constructor. */
		public FixBase(String description, Consumer<IConstraintInstanceStatus> fix) {
			this.description = description;
			this.fix = fix;
		}

		/** {@inheritDoc} */
		@Override
		public String getDescription() {
			return description;
		}

		/** {@inheritDoc} */
		@Override
		public void runFix(IConstraintInstanceStatus status) {
			fix.accept(status);
		}
	}
}
